#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.concept.proto import concept_proto_builder
from grakn.concept.type.type import Type, RemoteType


class ThingType(Type):

    def as_remote(self, transaction):
        return RemoteThingType(transaction, self.get_label(), self.is_root())

    def is_thing_type(self):
        return True


class RemoteThingType(RemoteType):

    def get_instances(self):
        method = concept_proto.Type.Req()
        get_instances_req = concept_proto.ThingType.GetInstances.Req()
        method.thing_type_get_instances_req.CopyFrom(get_instances_req)
        return self._thing_stream(method, lambda res: res.thing_type_get_instances_res.things)

    def set_abstract(self):
        req = concept_proto.Type.Req()
        req.thing_type_set_abstract_req.CopyFrom(concept_proto.ThingType.SetAbstract.Req())
        self._execute(req)

    def unset_abstract(self):
        req = concept_proto.Type.Req()
        req.thing_type_unset_abstract_req.CopyFrom(concept_proto.ThingType.UnsetAbstract.Req())
        self._execute(req)

    def set_plays(self, role, overridden_role=None):
        req = concept_proto.Type.Req()
        set_plays_req = concept_proto.ThingType.SetPlays.Req()
        set_plays_req.role.CopyFrom(concept_proto_builder.type_(role))
        if overridden_role:
            set_plays_req.overridden_role.CopyFrom(concept_proto_builder.type_(overridden_role))
        req.thing_type_set_plays_req.CopyFrom(set_plays_req)
        self._execute(req)

    def set_owns(self, attribute_type, overridden_type=None, is_key=False):
        req = concept_proto.Type.Req()
        set_owns_req = concept_proto.ThingType.SetOwns.Req()
        set_owns_req.attribute_type.CopyFrom(concept_proto_builder.type_(attribute_type))
        set_owns_req.is_key = is_key
        if overridden_type:
            set_owns_req.overridden_type.CopyFrom(concept_proto_builder.type_(overridden_type))
        req.thing_type_set_owns_req.CopyFrom(set_owns_req)
        self._execute(req)

    def get_plays(self):
        req = concept_proto.Type.Req()
        req.thing_type_get_plays_req.CopyFrom(concept_proto.ThingType.GetPlays.Req())
        return self._type_stream(req, lambda res: res.thing_type_get_plays_res.roles)

    def get_owns(self, value_type=None, keys_only=False):
        req = concept_proto.Type.Req()
        get_owns_req = concept_proto.ThingType.GetOwns.Req()
        get_owns_req.keys_only = keys_only
        if value_type:
            get_owns_req.value_type = concept_proto_builder.value_type(value_type)
        req.thing_type_get_owns_req.CopyFrom(get_owns_req)
        return self._type_stream(req, lambda res: res.thing_type_get_owns_res.attribute_types)

    def unset_plays(self, role):
        req = concept_proto.Type.Req()
        unset_plays_req = concept_proto.ThingType.UnsetPlays.Req()
        unset_plays_req.role.CopyFrom(concept_proto_builder.type_(role))
        req.thing_type_unset_plays_req.CopyFrom(unset_plays_req)
        self._execute(req)

    def unset_owns(self, attribute_type):
        req = concept_proto.Type.Req()
        unset_owns_req = concept_proto.ThingType.UnsetOwns.Req()
        unset_owns_req.attribute_type.CopyFrom(concept_proto_builder.type_(attribute_type))
        req.thing_type_unset_owns_req.CopyFrom(unset_owns_req)
        self._execute(req)

    def as_remote(self, transaction):
        return RemoteThingType(transaction, self.get_label(), self.is_root())

    def is_thing_type(self):
        return True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Callable, List

import grakn_protocol.protobuf.concept_pb2 as concept_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.common.exception import GraknClientException
from grakn.concept.proto import concept_proto_reader, concept_proto_builder
from grakn.concept.concept import Concept, RemoteConcept


class Thing(Concept):

    def __init__(self, iid: str):
        if not iid:
            raise GraknClientException("IID must be a non-empty string.")
        self._iid = iid
        self._hash = hash(iid)

    def get_iid(self):
        return self._iid

    def is_thing(self):
        return True

    def __str__(self):
        return type(self).__name__ + "[iid:" + self.get_iid() + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self.get_iid() == other.get_iid()

    def __hash__(self):
        return self._hash


class RemoteThing(RemoteConcept):

    def __init__(self, transaction, iid: str):
        if not transaction:
            raise GraknClientException("Transaction must be set.")
        if not iid:
            raise GraknClientException("IID must be set.")
        self._transaction = transaction
        self._iid = iid
        self._hash = hash(iid)

    def get_iid(self):
        return self._iid

    def get_type(self):
        method = concept_proto.Thing.Req()
        method.thing_get_type_req.CopyFrom(concept_proto.Thing.GetType.Req())
        return concept_proto_reader.type_(self._execute(method).thing_get_type_res.thing_type)

    def is_inferred(self):
        req = concept_proto.Thing.Req()
        req.thing_is_inferred_req.CopyFrom(concept_proto.Thing.IsInferred.Req())
        return self._execute(req).thing_is_inferred_res.inferred

    def get_has(self, attribute_type=None, attribute_types: List = None, only_key=False):
        if [bool(attribute_type), bool(attribute_types), only_key].count(True) > 1:
            raise GraknClientException("Only one filter can be applied at a time to get_has."
                                       "The possible filters are: [attribute_type, attribute_types, only_key]")
        if attribute_type:
            attribute_types = [attribute_type]
        method = concept_proto.Thing.Req()
        get_has_req = concept_proto.Thing.GetHas.Req()
        if only_key:
            get_has_req.keys_only = only_key
        elif attribute_types:
            get_has_req.attribute_types.extend(concept_proto_builder.types(attribute_types))
        method.thing_get_has_req.CopyFrom(get_has_req)
        return self._thing_stream(method, lambda res: res.thing_get_has_res.attributes)

    def get_plays(self):
        req = concept_proto.Thing.Req()
        req.thing_get_plays_req.CopyFrom(concept_proto.Thing.GetPlays.Req())
        return self._type_stream(req, lambda res: res.thing_get_plays_res.role_types)

    def get_relations(self, role_types: list = None):
        if not role_types:
            role_types = []
        method = concept_proto.Thing.Req()
        get_relations_req = concept_proto.Thing.GetRelations.Req()
        get_relations_req.role_types.extend(concept_proto_builder.types(role_types))
        method.thing_get_relations_req.CopyFrom(get_relations_req)
        return self._thing_stream(method, lambda res: res.thing_get_relations_res.relations)

    def set_has(self, attribute):
        method = concept_proto.Thing.Req()
        set_has_req = concept_proto.Thing.SetHas.Req()
        set_has_req.attribute.CopyFrom(concept_proto_builder.thing(attribute))
        method.thing_set_has_req.CopyFrom(set_has_req)
        self._execute(method)

    def unset_has(self, attribute):
        method = concept_proto.Thing.Req()
        unset_has_req = concept_proto.Thing.UnsetHas.Req()
        unset_has_req.attribute.CopyFrom(concept_proto_builder.thing(attribute))
        method.thing_unset_has_req.CopyFrom(unset_has_req)
        self._execute(method)

    def delete(self):
        method = concept_proto.Thing.Req()
        method.thing_delete_req.CopyFrom(concept_proto.Thing.Delete.Req())
        self._execute(method)

    def is_deleted(self):
        return not self._transaction.concepts().get_thing(self.get_iid())

    def is_thing(self):
        return True

    def _thing_stream(self, method: concept_proto.Thing.Req, thing_list_getter: Callable[[concept_proto.Thing.Res], List[concept_proto.Thing]]):
        method.iid = concept_proto_builder.iid(self.get_iid())
        request = transaction_proto.Transaction.Req()
        request.thing_req.CopyFrom(method)
        return map(lambda thing_proto: concept_proto_reader.thing(thing_proto), self._transaction._stream(request, lambda res: thing_list_getter(res.thing_res)))

    def _type_stream(self, method: concept_proto.Thing.Req, type_list_getter: Callable[[concept_proto.Thing.Res], List[concept_proto.Type]]):
        method.iid = concept_proto_builder.iid(self.get_iid())
        request = transaction_proto.Transaction.Req()
        request.thing_req.CopyFrom(method)
        return map(lambda type_proto: concept_proto_reader.type_(type_proto), self._transaction._stream(request, lambda res: type_list_getter(res.thing_res)))

    def _execute(self, method: concept_proto.Thing.Req):
        method.iid = concept_proto_builder.iid(self.get_iid())
        request = transaction_proto.Transaction.Req()
        request.thing_req.CopyFrom(method)
        return self._transaction._execute(request).thing_res

    def __str__(self):
        return type(self).__name__ + "[iid:" + str(self._iid) + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self._transaction is other._transaction and self._iid == other._iid

    def __hash__(self):
        return self._hash



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Callable, List

import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.common.exception import GraknClientException


class Stream:

    _CONTINUE = "continue"
    _DONE = "done"

    def __init__(self, transaction, request_id: str, transform_response: Callable[[transaction_proto.Transaction.Res], List] = lambda res: None):
        self._transaction = transaction
        self._request_id = request_id
        self._transform_response = transform_response
        self._current_iterator = None
        self._done = False

    def __iter__(self):
        return self

    def __next__(self):
        if self._done:
            raise StopIteration()
        if self._current_iterator is not None:
            try:
                return next(self._current_iterator)
            except StopIteration:
                self._current_iterator = None

        res = self._transaction._fetch(self._request_id)
        res_case = res.WhichOneof("res")
        if res_case == Stream._CONTINUE:
            continue_req = transaction_proto.Transaction.Req()
            continue_req.id = self._request_id
            setattr(continue_req, "continue", True)
            self._transaction._request_iterator.put(continue_req)
            return next(self)
        elif res_case == Stream._DONE:
            self._done = True
            raise StopIteration()
        elif res_case is None:
            raise GraknClientException("The required field 'res' of type 'transaction_proto.Transaction.Res' was not set.")
        else:
            self._current_iterator = iter(self._transform_response(res))
            return next(self)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import ABC, abstractmethod
from typing import List, Set, Dict

import grakn_protocol.protobuf.cluster.cluster_pb2 as cluster_proto
import grpc
from grakn_protocol.protobuf.cluster.grakn_cluster_pb2_grpc import GraknClusterStub
from grpc import RpcError

# Repackaging these symbols allows them to be imported from "grakn.client"
from grakn.common.exception import GraknClientException  # noqa # pylint: disable=unused-import
from grakn.concept.type.value_type import ValueType  # noqa # pylint: disable=unused-import
from grakn.options import GraknOptions, GraknClusterOptions
from grakn.rpc.cluster.failsafe_task import _FailsafeTask
from grakn.rpc.cluster.database import DatabaseCluster, _DatabaseClusterRPC  # noqa # pylint: disable=unused-import
from grakn.rpc.cluster.server_address import ServerAddress
from grakn.rpc.cluster.database_manager import DatabaseManagerCluster, _DatabaseManagerClusterRPC
from grakn.rpc.cluster.session import SessionClusterRPC
from grakn.rpc.database_manager import DatabaseManager, _DatabaseManagerRPC
from grakn.rpc.session import Session, SessionType, _SessionRPC
from grakn.rpc.transaction import TransactionType  # noqa # pylint: disable=unused-import


class GraknClient(ABC):
    DEFAULT_ADDRESS = "localhost:1729"

    @staticmethod
    def core(address=DEFAULT_ADDRESS) -> "GraknClient":
        return _ClientRPC(address)

    @staticmethod
    def cluster(addresses: List[str]) -> "GraknClientCluster":
        return _ClientClusterRPC(addresses)

    @abstractmethod
    def session(self, database: str, session_type: SessionType, options: GraknOptions = None) -> Session:
        pass

    @abstractmethod
    def databases(self) -> DatabaseManager:
        pass

    @abstractmethod
    def is_open(self) -> bool:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def is_cluster(self) -> bool:
        pass

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class GraknClientCluster(GraknClient):

    @abstractmethod
    def databases(self) -> DatabaseManagerCluster:
        pass


class _ClientRPC(GraknClient):

    def __init__(self, address: str):
        self._address = address
        self._channel = grpc.insecure_channel(self._address)
        self._databases = _DatabaseManagerRPC(self._channel)
        self._sessions: Dict[bytes, _SessionRPC] = {}
        self._is_open = True

    def session(self, database: str, session_type: SessionType, options=None) -> Session:
        if not options:
            options = GraknOptions.core()
        session = _SessionRPC(self, database, session_type, options)
        self._sessions[session.session_id()] = session
        return session

    def databases(self):
        return self._databases

    def is_open(self):
        return self._is_open

    def close(self):
        for session_id in self._sessions:
            self._sessions[session_id].close()
        self._channel.close()
        self._is_open = False

    def is_cluster(self) -> bool:
        return False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_tb is None:
            pass
        else:
            return False

    def remove_session(self, session: _SessionRPC):
        del self._sessions[session.session_id()]

    def channel(self):
        return self._channel


# _ClientClusterRPC must live in this package because of circular ref with GraknClient
class _ClientClusterRPC(GraknClientCluster):

    def __init__(self, addresses: List[str]):
        self._core_clients: Dict[ServerAddress, _ClientRPC] = {addr: _ClientRPC(addr.client()) for addr in self._fetch_cluster_servers(addresses)}
        self._grakn_cluster_grpc_stubs = {addr: GraknClusterStub(client.channel()) for (addr, client) in self._core_clients.items()}
        self._database_managers = _DatabaseManagerClusterRPC(self, {addr: client.databases() for (addr, client) in self._core_clients.items()})
        self._cluster_databases: Dict[str, _DatabaseClusterRPC] = {}
        self._is_open = True

    def session(self, database: str, session_type: SessionType, options=None) -> Session:
        if not options:
            options = GraknOptions.cluster()
        return self._session_any_replica(database, session_type, options) if options.read_any_replica else self._session_primary_replica(database, session_type, options)

    def _session_primary_replica(self, database: str, session_type: SessionType, options=None) -> SessionClusterRPC:
        return _OpenSessionFailsafeTask(database, session_type, options, self).run_primary_replica()

    def _session_any_replica(self, database: str, session_type: SessionType, options=None) -> SessionClusterRPC:
        return _OpenSessionFailsafeTask(database, session_type, options, self).run_any_replica()

    def databases(self) -> DatabaseManagerCluster:
        return self._database_managers

    def is_open(self) -> bool:
        return self._is_open

    def close(self) -> None:
        for client in self._core_clients.values():
            client.close()
        self._is_open = False

    def is_cluster(self) -> bool:
        return True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def cluster_databases(self) -> Dict[str, _DatabaseClusterRPC]:
        return self._cluster_databases

    def cluster_members(self) -> Set[ServerAddress]:
        return set(self._core_clients.keys())

    def core_client(self, address: ServerAddress) -> _ClientRPC:
        return self._core_clients.get(address)

    def grakn_cluster_grpc_stub(self, address: ServerAddress) -> GraknClusterStub:
        return self._grakn_cluster_grpc_stubs.get(address)

    def _fetch_cluster_servers(self, addresses: List[str]) -> Set[ServerAddress]:
        for address in addresses:
            try:
                with _ClientRPC(address) as client:
                    print("Performing cluster discovery to %s..." % address)
                    grakn_cluster_stub = GraknClusterStub(client.channel())
                    res = grakn_cluster_stub.cluster_servers(cluster_proto.Cluster.Servers.Req())
                    members = set([ServerAddress.parse(srv) for srv in res.servers])
                    print("Discovered %s" % [str(member) for member in members])
                    return members
            except RpcError as e:
                print("Cluster discovery to %s failed. %s" % (address, str(e)))
        raise GraknClientException("Unable to connect to Grakn Cluster. Attempted connecting to the cluster members, but none are available: %s" % str(addresses))


class _OpenSessionFailsafeTask(_FailsafeTask):

    def __init__(self, database: str, session_type: SessionType, options: GraknClusterOptions, client: "_ClientClusterRPC"):
        super().__init__(client, database)
        self.session_type = session_type
        self.options = options

    def run(self, replica: _DatabaseClusterRPC.Replica):
        return SessionClusterRPC(self.client, replica.address(), self.database, self.session_type, self.options)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Callable, List

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.concept.proto import concept_proto_reader
from grakn.concept.type.type import Type, RemoteType


class RoleType(Type):

    def __init__(self, label: str, scope: str, is_root: bool):
        super(RoleType, self).__init__(label, is_root)
        self._scope = scope
        self._hash = hash((scope, label))

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return RoleType(type_proto.label, type_proto.scope, type_proto.root)

    def get_scope(self):
        return self._scope

    def get_scoped_label(self):
        return self.get_scope() + ":" + self.get_label()

    def as_remote(self, transaction):
        return RemoteRoleType(transaction, self.get_label(), self.get_scope(), self.is_root())

    def is_role_type(self):
        return True

    def __str__(self):
        return type(self).__name__ + "[label:" + self.get_scoped_label() + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self.get_scoped_label() == other.get_scoped_label()

    def __hash__(self):
        return super(RoleType, self).__hash__()


class RemoteRoleType(RemoteType):

    def __init__(self, transaction, label: str, scope: str, is_root: bool):
        super(RemoteRoleType, self).__init__(transaction, label, is_root)
        self._scope = scope
        self._hash = hash((transaction, scope, label))

    def get_scope(self):
        return self._scope

    def get_scoped_label(self):
        return self.get_scope() + ":" + self.get_label()

    def as_remote(self, transaction):
        return RemoteRoleType(transaction, self.get_label(), self.get_scope(), self.is_root())

    def get_relation_type(self):
        method = concept_proto.Type.Req()
        method.role_type_get_relation_type_req.CopyFrom(concept_proto.RoleType.GetRelationType.Req())
        return concept_proto_reader.type_(self._execute(method).role_type_get_relation_type_res.relation_type)

    def get_relation_types(self):
        method = concept_proto.Type.Req()
        method.role_type_get_relation_types_req.CopyFrom(concept_proto.RoleType.GetRelationTypes.Req())
        return self._type_stream(method, lambda res: res.role_type_get_relation_types_res.relation_types)

    def get_players(self):
        method = concept_proto.Type.Req()
        method.role_type_get_players_req.CopyFrom(concept_proto.RoleType.GetPlayers.Req())
        return self._type_stream(method, lambda res: res.role_type_get_players_res.thing_types)

    def is_role_type(self):
        return True

    def _type_stream(self, method: concept_proto.Type.Req, type_list_getter: Callable[[concept_proto.Type.Res], List[concept_proto.Type]]):
        method.scope = self.get_scope()
        return super(RemoteRoleType, self)._type_stream(method, type_list_getter)

    def _execute(self, method: concept_proto.Type.Req):
        method.scope = self.get_scope()
        return super(RemoteRoleType, self)._execute(method)

    def __str__(self):
        return type(self).__name__ + "[label:" + self.get_scoped_label() + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self.get_scoped_label() == other.get_scoped_label()

    def __hash__(self):
        return super(RemoteRoleType, self).__hash__()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.common.exception import GraknClientException
from grakn.concept.answer.concept_map import ConceptMap
from grakn.concept.thing.attribute import BooleanAttribute, LongAttribute, DoubleAttribute, StringAttribute, \
    DateTimeAttribute
from grakn.concept.thing.entity import Entity
from grakn.concept.thing.relation import Relation
from grakn.concept.type.attribute_type import BooleanAttributeType, LongAttributeType, DoubleAttributeType, \
    StringAttributeType, DateTimeAttributeType, AttributeType
from grakn.concept.type.entity_type import EntityType
from grakn.concept.type.relation_type import RelationType
from grakn.concept.type.role_type import RoleType
from grakn.concept.type.thing_type import ThingType


def iid(iid_proto: bytes):
    return "0x" + iid_proto.hex()


def concept(con_proto: concept_proto.Concept):
    if con_proto.HasField(ConceptMap._THING):
        concept = thing(con_proto.thing)
    else:
        concept = type_(con_proto.type)
    return concept


def thing(thing_proto: concept_proto.Thing):
    if thing_proto.encoding == concept_proto.Thing.Encoding.Value("ENTITY"):
        return Entity._of(thing_proto)
    elif thing_proto.encoding == concept_proto.Thing.Encoding.Value("RELATION"):
        return Relation._of(thing_proto)
    elif thing_proto.encoding == concept_proto.Thing.Encoding.Value("ATTRIBUTE"):
        return attribute(thing_proto)
    else:
        raise GraknClientException("The encoding " + thing_proto.encoding + " was not recognised.")


def attribute(thing_proto: concept_proto.Thing):
    if thing_proto.value_type == concept_proto.AttributeType.ValueType.Value("BOOLEAN"):
        return BooleanAttribute._of(thing_proto)
    elif thing_proto.value_type == concept_proto.AttributeType.ValueType.Value("LONG"):
        return LongAttribute._of(thing_proto)
    elif thing_proto.value_type == concept_proto.AttributeType.ValueType.Value("DOUBLE"):
        return DoubleAttribute._of(thing_proto)
    elif thing_proto.value_type == concept_proto.AttributeType.ValueType.Value("STRING"):
        return StringAttribute._of(thing_proto)
    elif thing_proto.value_type == concept_proto.AttributeType.ValueType.Value("DATETIME"):
        return DateTimeAttribute._of(thing_proto)
    else:
        raise GraknClientException("The value type " + str(thing_proto.value_type) + " was not recognised.")


def type_(type_proto: concept_proto.Type):
    if type_proto.encoding == concept_proto.Type.Encoding.Value("ROLE_TYPE"):
        return RoleType._of(type_proto)
    else:
        return thing_type(type_proto)


def thing_type(type_proto: concept_proto.Type):
    if type_proto.encoding == concept_proto.Type.Encoding.Value("ENTITY_TYPE"):
        return EntityType._of(type_proto)
    elif type_proto.encoding == concept_proto.Type.Encoding.Value("RELATION_TYPE"):
        return RelationType._of(type_proto)
    elif type_proto.encoding == concept_proto.Type.Encoding.Value("ATTRIBUTE_TYPE"):
        return attribute_type(type_proto)
    elif type_proto.encoding == concept_proto.Type.Encoding.Value("THING_TYPE"):
        return ThingType(type_proto.label, type_proto.root)
    else:
        raise GraknClientException("The encoding " + str(type_proto.encoding) + " was not recognised.")


def attribute_type(type_proto: concept_proto.Type):
    if type_proto.value_type == concept_proto.AttributeType.ValueType.Value("BOOLEAN"):
        return BooleanAttributeType._of(type_proto)
    elif type_proto.value_type == concept_proto.AttributeType.ValueType.Value("LONG"):
        return LongAttributeType._of(type_proto)
    elif type_proto.value_type == concept_proto.AttributeType.ValueType.Value("DOUBLE"):
        return DoubleAttributeType._of(type_proto)
    elif type_proto.value_type == concept_proto.AttributeType.ValueType.Value("STRING"):
        return StringAttributeType._of(type_proto)
    elif type_proto.value_type == concept_proto.AttributeType.ValueType.Value("DATETIME"):
        return DateTimeAttributeType._of(type_proto)
    elif type_proto.value_type == concept_proto.AttributeType.ValueType.Value("OBJECT"):
        return AttributeType(type_proto.label, type_proto.root)
    else:
        raise GraknClientException("The value type " + str(type_proto.value_type) + " was not recognised.")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from grakn.common.exception import GraknClientException


class ServerAddress:

    def __init__(self, client_host: str, client_port: int, server_host: str, server_port: int):
        self._client_host = client_host
        self._client_port = client_port
        self._server_host = server_host
        self._server_port = server_port

    def client_host(self) -> str:
        return self._client_host

    def client_port(self) -> int:
        return self._client_port

    def server_host(self) -> str:
        return self._server_host

    def server_port(self) -> int:
        return self._server_port

    def server(self) -> str:
        return "%s:%d" % (self._server_host, self._server_port)

    def client(self) -> str:
        return "%s:%d" % (self._client_host, self._client_port)

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self._client_host == other.client_host() and self._client_port == other.client_port() and self._server_host == other.server_host() and self._server_port == other.server_port()

    def __hash__(self):
        return hash((self._client_host, self._client_port, self._server_host, self._server_port))

    def __str__(self):
        return "%s,%s" % (self.client(), self.server())

    @staticmethod
    def parse(address: str) -> "ServerAddress":
        s = address.split(",")
        if len(s) == 1:
            s1 = address.split(":")
            return ServerAddress(client_host=s1[0], client_port=int(s1[1]), server_host=s1[0], server_port=int(s1[1]))
        elif len(s) == 2:
            client_url = s[0].split(":")
            server_url = s[1].split(":")
            if len(client_url) != 2 or len(server_url) != 2:
                raise GraknClientException("Failed to parse server address: " + address)
            return ServerAddress(client_host=client_url[0], client_port=int(client_url[1]), server_host=server_url[0], server_port=int(server_url[1]))
        else:
            raise GraknClientException("Failed to parse server address: " + address)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.concept_pb2 as concept_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.concept.proto import concept_proto_builder, concept_proto_reader
from grakn.concept.thing.thing import Thing, RemoteThing


class Relation(Thing):

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return Relation(concept_proto_reader.iid(thing_proto.iid))

    def as_remote(self, transaction):
        return RemoteRelation(transaction, self.get_iid())

    def is_relation(self):
        return True


class RemoteRelation(RemoteThing):

    def as_remote(self, transaction):
        return RemoteRelation(transaction, self.get_iid())

    def get_players_by_role_type(self):
        method = concept_proto.Thing.Req()
        method.relation_get_players_by_role_type_req.CopyFrom(concept_proto.Relation.GetPlayersByRoleType.Req())
        method.iid = concept_proto_builder.iid(self.get_iid())

        request = transaction_proto.Transaction.Req()
        request.thing_req.CopyFrom(method)
        stream = self._transaction._stream(request, lambda res: res.thing_res.relation_get_players_by_role_type_res.role_types_with_players)

        role_player_dict = {}
        for role_player in stream:
            role = concept_proto_reader.type_(role_player.role_type)
            player = concept_proto_reader.thing(role_player.player)
            if role not in role_player_dict:
                role_player_dict[role] = []
            role_player_dict[role].append(player)
        return role_player_dict

    def get_players(self, role_types=None):
        if not role_types:
            role_types = []
        method = concept_proto.Thing.Req()
        get_players_req = concept_proto.Relation.GetPlayers.Req()
        get_players_req.role_types.extend(concept_proto_builder.types(role_types))
        method.relation_get_players_req.CopyFrom(get_players_req)
        return self._thing_stream(method, lambda res: res.relation_get_players_res.things)

    def add_player(self, role_type, player):
        method = concept_proto.Thing.Req()
        add_player_req = concept_proto.Relation.AddPlayer.Req()
        add_player_req.role_type.CopyFrom(concept_proto_builder.type_(role_type))
        add_player_req.player.CopyFrom(concept_proto_builder.thing(player))
        method.relation_add_player_req.CopyFrom(add_player_req)
        self._execute(method)

    def remove_player(self, role_type, player):
        method = concept_proto.Thing.Req()
        remove_player_req = concept_proto.Relation.RemovePlayer.Req()
        remove_player_req.role_type.CopyFrom(concept_proto_builder.type_(role_type))
        remove_player_req.player.CopyFrom(concept_proto_builder.thing(player))
        method.relation_remove_player_req.CopyFrom(remove_player_req)
        self._execute(method)

    def is_relation(self):
        return True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import enum
from typing import Callable, List

import grpc
import time
import uuid

import queue

from grakn_protocol.protobuf.grakn_pb2_grpc import GraknStub
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn import grakn_proto_builder
from grakn.common.exception import GraknClientException
from grakn.concept.concept_manager import ConceptManager
from grakn.options import GraknOptions
from grakn.query.query_manager import QueryManager
from grakn.rpc.stream import Stream
from grakn.logic.logic_manager import LogicManager


class TransactionType(enum.Enum):
    READ = 0
    WRITE = 1


class Transaction:

    def __init__(self, address: str, session_id: bytes, transaction_type: TransactionType, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        self._transaction_type = transaction_type
        self._concept_manager = ConceptManager(self)
        self._query_manager = QueryManager(self)
        self._logic_manager = LogicManager(self)
        self._response_queues = {}

        self._grpc_stub = GraknStub(grpc.insecure_channel(address))
        self._request_iterator = RequestIterator()
        self._response_iterator = self._grpc_stub.transaction(self._request_iterator)
        self._transaction_was_closed = False

        open_req = transaction_proto.Transaction.Open.Req()
        open_req.session_id = session_id
        open_req.type = Transaction._transaction_type_proto(transaction_type)
        open_req.options.CopyFrom(grakn_proto_builder.options(options))
        req = transaction_proto.Transaction.Req()
        req.open_req.CopyFrom(open_req)

        start_time = time.time() * 1000.0
        res = self._execute(req)
        end_time = time.time() * 1000.0
        self._network_latency_millis = end_time - start_time - res.open_res.processing_time_millis

    def transaction_type(self):
        return self._transaction_type

    def is_open(self):
        return not self._transaction_was_closed

    def concepts(self):
        return self._concept_manager

    def query(self):
        return self._query_manager

    def logic(self):
        return self._logic_manager

    def commit(self):
        req = transaction_proto.Transaction.Req()
        commit_req = transaction_proto.Transaction.Commit.Req()
        req.commit_req.CopyFrom(commit_req)
        try:
            self._execute(req)
        finally:
            self.close()

    def rollback(self):
        req = transaction_proto.Transaction.Req()
        rollback_req = transaction_proto.Transaction.Rollback.Req()
        req.rollback_req.CopyFrom(rollback_req)
        self._execute(req)

    def close(self):
        self._transaction_was_closed = True
        self._request_iterator.close()

    def _execute(self, request: transaction_proto.Transaction.Req):
        response_queue = queue.Queue()
        request_id = str(uuid.uuid4())
        request.id = request_id
        if self._transaction_was_closed:
            raise GraknClientException("The transaction has been closed and no further operation is allowed.")
        self._response_queues[request_id] = response_queue
        self._request_iterator.put(request)
        return self._fetch(request_id)

    def _stream(self, request: transaction_proto.Transaction.Req, transform_response: Callable[[transaction_proto.Transaction.Res], List] = None):
        response_queue = queue.Queue()
        request_id = str(uuid.uuid4())
        request.id = request_id
        if self._transaction_was_closed:
            raise GraknClientException("The transaction has been closed and no further operation is allowed.")
        self._response_queues[request_id] = response_queue
        self._request_iterator.put(request)
        return Stream(self, request_id, transform_response)

    def _fetch(self, request_id: str):
        try:
            return self._response_queues[request_id].get(block=False)
        except queue.Empty:
            pass

        # Keep taking responses until we get one that matches the request ID
        while True:
            try:
                response = next(self._response_iterator)
            except grpc.RpcError as e:
                self._transaction_was_closed = True
                grakn_exception = GraknClientException(e.details())
                for response_queue in self._response_queues.values():
                    response_queue.put(grakn_exception)
                # noinspection PyUnresolvedReferences
                raise grakn_exception
            except StopIteration:
                raise GraknClientException("The transaction has been closed and no further operation is allowed.")

            if isinstance(response, GraknClientException):
                raise response
            elif response.id == request_id:
                return response
            else:
                response_queue = self._response_queues[response.id]
                if response_queue is None:
                    raise GraknClientException("Received a response with unknown request id '" + response.id + "'.")
                response_queue.put(response)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_tb is None:
            pass
        else:
            return False

    @staticmethod
    def _transaction_type_proto(transaction_type):
        if transaction_type == TransactionType.READ:
            return transaction_proto.Transaction.Type.Value("READ")
        if transaction_type == TransactionType.WRITE:
            return transaction_proto.Transaction.Type.Value("WRITE")


class RequestIterator:
    CLOSE_STREAM = "CLOSE_STREAM"

    def __init__(self):
        self._request_queue = queue.Queue()

    def __iter__(self):
        return self

    # Essentially the gRPC stream is constantly polling this iterator. When we issue a new request, it gets put into
    # the back of the queue and gRPC will pick it up when it gets round to it (this is usually instantaneous)
    def __next__(self):
        request = self._request_queue.get(block=True)
        if request is RequestIterator.CLOSE_STREAM:
            # Close the stream.
            raise StopIteration()
        return request

    def put(self, request: transaction_proto.Transaction.Req):
        self._request_queue.put(request)

    def close(self):
        self._request_queue.put(RequestIterator.CLOSE_STREAM)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Callable, List

import grakn_protocol.protobuf.concept_pb2 as concept_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.common.exception import GraknClientException
from grakn.concept.proto import concept_proto_builder, concept_proto_reader
from grakn.concept.concept import Concept, RemoteConcept


class Type(Concept):

    def __init__(self, label: str, is_root: bool):
        if not label:
            raise GraknClientException("Label must be a non-empty string.")
        self._label = label
        self._is_root = is_root
        self._hash = hash(label)

    def get_label(self):
        return self._label

    def is_root(self):
        return self._is_root

    def is_type(self):
        return True

    def __str__(self):
        return type(self).__name__ + "[label:" + self.get_label() + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self.get_label() == other.get_label()

    def __hash__(self):
        return self._hash


class RemoteType(RemoteConcept):

    def __init__(self, transaction, label: str, is_root: bool):
        if not transaction:
            raise GraknClientException("Transaction must be set.")
        if not label:
            raise GraknClientException("Label must be a non-empty string.")
        self._transaction = transaction
        self._label = label
        self._is_root = is_root
        self._hash = hash((self._transaction, label))

    def get_label(self):
        return self._label

    def is_root(self):
        return self._is_root

    def set_label(self, label: str):
        req = concept_proto.Type.Req()
        set_label_req = concept_proto.Type.SetLabel.Req()
        set_label_req.label = label
        req.type_set_label_req.CopyFrom(set_label_req)
        self._execute(req)
        self._label = label
        self._hash = hash((self._transaction, label))

    def is_abstract(self):
        req = concept_proto.Type.Req()
        req.type_is_abstract_req.CopyFrom(concept_proto.Type.IsAbstract.Req())
        res = self._execute(req)
        return res.type_is_abstract_res.abstract

    def is_type(self):
        return True

    def set_supertype(self, _type: Type):
        req = concept_proto.Type.Req()
        supertype_req = concept_proto.Type.SetSupertype.Req()
        supertype_req.type.CopyFrom(concept_proto_builder.type_(_type))
        req.type_set_supertype_req.CopyFrom(supertype_req)
        self._execute(req)

    def get_supertype(self):
        req = concept_proto.Type.Req()
        req.type_get_supertype_req.CopyFrom(concept_proto.Type.GetSupertype.Req())
        res = self._execute(req).type_get_supertype_res
        return concept_proto_reader.type_(res.type) if res.WhichOneof("res") == "type" else None

    def get_supertypes(self):
        method = concept_proto.Type.Req()
        method.type_get_supertypes_req.CopyFrom(concept_proto.Type.GetSupertypes.Req())
        return self._type_stream(method, lambda res: res.type_get_supertypes_res.types)

    def get_subtypes(self):
        method = concept_proto.Type.Req()
        method.type_get_subtypes_req.CopyFrom(concept_proto.Type.GetSubtypes.Req())
        return self._type_stream(method, lambda res: res.type_get_subtypes_res.types)

    def delete(self):
        method = concept_proto.Type.Req()
        method.type_delete_req.CopyFrom(concept_proto.Type.Delete.Req())
        self._execute(method)

    def is_deleted(self):
        return not self._transaction.concepts()._get_thing_type(self.get_label())

    def _type_stream(self, method: concept_proto.Type.Req, type_list_getter: Callable[[concept_proto.Type.Res], List[concept_proto.Type]]):
        method.label = self.get_label()
        request = transaction_proto.Transaction.Req()
        request.type_req.CopyFrom(method)
        return map(lambda type_proto: concept_proto_reader.type_(type_proto), self._transaction._stream(request, lambda res: type_list_getter(res.type_res)))

    def _thing_stream(self, method: concept_proto.Type.Req, thing_list_getter: Callable[[concept_proto.Type.Res], List[concept_proto.Thing]]):
        method.label = self.get_label()
        request = transaction_proto.Transaction.Req()
        request.type_req.CopyFrom(method)
        return map(lambda thing_proto: concept_proto_reader.thing(thing_proto), self._transaction._stream(request, lambda res: thing_list_getter(res.type_res)))

    def _execute(self, method: concept_proto.Type.Req):
        method.label = self.get_label()
        request = transaction_proto.Transaction.Req()
        request.type_req.CopyFrom(method)
        return self._transaction._execute(request).type_res

    def __str__(self):
        return type(self).__name__ + "[label:" + self.get_label() + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self._transaction is other._transaction and self.get_label() == other.get_label()

    def __hash__(self):
        return self._hash



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import ABC, abstractmethod
from typing import List

import grakn_protocol.protobuf.database_pb2 as database_proto
from grakn_protocol.protobuf.grakn_pb2_grpc import GraknStub
from grpc import Channel

from grakn.common.exception import GraknClientException
from grakn.rpc.database import Database, _DatabaseRPC
from grakn.rpc.utils import rpc_call


class DatabaseManager(ABC):

    @abstractmethod
    def contains(self, name: str) -> bool:
        pass

    @abstractmethod
    def create(self, name: str) -> None:
        pass

    @abstractmethod
    def get(self, name: str) -> Database:
        pass

    @abstractmethod
    def all(self) -> List[Database]:
        pass


def _not_blank(name: str) -> str:
    if name in [None, ""] or name.isspace():
        raise GraknClientException("Database name must not be empty.")
    return name


class _DatabaseManagerRPC(DatabaseManager):

    def __init__(self, channel: Channel):
        self._grpc_stub = GraknStub(channel)

    def contains(self, name: str) -> bool:
        request = database_proto.Database.Contains.Req()
        request.name = _not_blank(name)
        return rpc_call(lambda: self._grpc_stub.database_contains(request).contains)

    def create(self, name: str) -> None:
        request = database_proto.Database.Create.Req()
        request.name = _not_blank(name)
        rpc_call(lambda: self._grpc_stub.database_create(request))

    def get(self, name: str) -> Database:
        if self.contains(name):
            return _DatabaseRPC(self, name)
        else:
            raise GraknClientException("The database '%s' does not exist." % name)

    def all(self) -> List[Database]:
        databases: List[str] = rpc_call(lambda: list(self._grpc_stub.database_all(database_proto.Database.All.Req()).names))
        return [_DatabaseRPC(self, name) for name in databases]

    def grpc_stub(self) -> GraknStub:
        return self._grpc_stub



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import ABC, abstractmethod
from typing import Set, Optional

import grakn_protocol.protobuf.database_pb2 as database_proto

from grakn.rpc.cluster.server_address import ServerAddress
from grakn.rpc.utils import rpc_call


class Database(ABC):

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def delete(self) -> None:
        pass

    class Replica(ABC):

        @abstractmethod
        def database(self) -> "DatabaseCluster":
            pass

        @abstractmethod
        def term(self) -> int:
            pass

        @abstractmethod
        def is_primary(self) -> bool:
            pass

        @abstractmethod
        def is_preferred_secondary(self) -> bool:
            pass

        @abstractmethod
        def address(self) -> ServerAddress:
            pass


class DatabaseCluster(Database, ABC):

    @abstractmethod
    def replicas(self) -> Set[Database.Replica]:
        pass

    @abstractmethod
    def primary_replica(self) -> Optional[Database.Replica]:
        pass

    @abstractmethod
    def preferred_secondary_replica(self) -> Database.Replica:
        pass


class _DatabaseRPC(Database):

    def __init__(self, database_manager, name: str):
        self._name = name
        self._grpc_stub = database_manager.grpc_stub()

    def name(self) -> str:
        return self._name

    def delete(self) -> None:
        request = database_proto.Database.Delete.Req()
        request.name = self._name
        rpc_call(lambda: self._grpc_stub.database_delete(request))

    def __str__(self):
        return self._name



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Mapping

import grakn_protocol.protobuf.answer_pb2 as answer_proto

from grakn.common.exception import GraknClientException
from grakn.concept.proto import concept_proto_reader
from grakn.concept.concept import Concept


class ConceptMap:

    _THING = "thing"

    def __init__(self, mapping: Mapping[str, Concept]):
        self._map = mapping

    def map(self):
        return self._map

    def concepts(self):
        return self._map.values()

    def get(self, variable: str):
        concept = self._map[variable]
        if not concept:
            raise GraknClientException("The variable " + variable + " does not exist.")
        return concept

    def __str__(self):
        return "".join(map(lambda var: "[" + var + "/" + str(self._map[var]) + "]", sorted(self._map.keys())))

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(other) != type(self):
            return False
        return other._map == self._map

    def __hash__(self):
        return hash(self._map)


def _of(concept_map_proto: answer_proto.ConceptMap):
    variable_map = {}
    for res_var in concept_map_proto.map:
        concept = concept_proto_reader.concept(concept_map_proto.map[res_var])
        variable_map[res_var] = concept
    return ConceptMap(variable_map)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Callable, List

import grakn_protocol.protobuf.query_pb2 as query_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn import grakn_proto_builder
from grakn.concept.answer import concept_map, concept_map_group, numeric, numeric_group
from grakn.options import GraknOptions


class QueryManager:

    def __init__(self, transaction):
        self._transaction = transaction

    def match(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        match_req = query_proto.Query.Match.Req()
        match_req.query = query
        request.match_req.CopyFrom(match_req)
        return map(lambda answer_proto: concept_map._of(answer_proto), self._iterate_query(request, lambda res: res.query_res.match_res.answers, options))

    def match_aggregate(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        match_aggregate_req = query_proto.Query.MatchAggregate.Req()
        match_aggregate_req.query = query
        request.match_aggregate_req.CopyFrom(match_aggregate_req)
        return self._iterate_query(request, lambda res: [numeric._of(res.query_res.match_aggregate_res.answer)], options)

    def match_group(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        match_group_req = query_proto.Query.MatchGroup.Req()
        match_group_req.query = query
        request.match_group_req.CopyFrom(match_group_req)
        return map(
            lambda cmg_proto: concept_map_group._of(cmg_proto),
            self._iterate_query(request, lambda res: res.query_res.match_group_res.answers, options)
        )

    def match_group_aggregate(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        match_group_aggregate_req = query_proto.Query.MatchGroupAggregate.Req()
        match_group_aggregate_req.query = query
        request.match_group_aggregate_req.CopyFrom(match_group_aggregate_req)
        return map(
            lambda numeric_group_proto: numeric_group._of(numeric_group_proto),
            self._iterate_query(request, lambda res: res.query_res.match_group_aggregate_res.answers, options)
        )

    def insert(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        insert_req = query_proto.Query.Insert.Req()
        insert_req.query = query
        request.insert_req.CopyFrom(insert_req)
        return map(lambda answer_proto: concept_map._of(answer_proto), self._iterate_query(request, lambda res: res.query_res.insert_res.answers, options))

    def delete(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        delete_req = query_proto.Query.Delete.Req()
        delete_req.query = query
        request.delete_req.CopyFrom(delete_req)
        return self._iterate_query(request, lambda res: [], options)

    def update(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        update_req = query_proto.Query.Update.Req()
        update_req.query = query
        request.update_req.CopyFrom(update_req)
        return map(lambda answer_proto: concept_map._of(answer_proto), self._iterate_query(request, lambda res: res.query_res.update_res.answers, options))

    def define(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        define_req = query_proto.Query.Define.Req()
        define_req.query = query
        request.define_req.CopyFrom(define_req)
        return self._iterate_query(request, lambda res: [], options)

    def undefine(self, query: str, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        request = query_proto.Query.Req()
        undefine_req = query_proto.Query.Undefine.Req()
        undefine_req.query = query
        request.undefine_req.CopyFrom(undefine_req)
        return self._iterate_query(request, lambda res: [], options)

    def _iterate_query(self, query_req: query_proto.Query.Req, response_reader: Callable[[transaction_proto.Transaction.Res], List], options: GraknOptions):
        req = transaction_proto.Transaction.Req()
        query_req.options.CopyFrom(grakn_proto_builder.options(options))
        req.query_req.CopyFrom(query_req)
        return self._transaction._stream(req, response_reader)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Callable, List

import grakn_protocol.protobuf.logic_pb2 as logic_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.logic.rule import Rule


class LogicManager:

    def __init__(self, transaction):
        self._transaction = transaction

    def put_rule(self, label: str, when: str, then: str):
        req = logic_proto.LogicManager.Req()
        put_rule_req = logic_proto.LogicManager.PutRule.Req()
        put_rule_req.label = label
        put_rule_req.when = when
        put_rule_req.then = then
        req.put_rule_req.CopyFrom(put_rule_req)
        res = self._execute(req)
        return Rule._of(res.put_rule_res.rule)

    def get_rule(self, label: str):
        req = logic_proto.LogicManager.Req()
        get_rule_req = logic_proto.LogicManager.GetRule.Req()
        get_rule_req.label = label
        req.get_rule_req.CopyFrom(get_rule_req)

        response = self._execute(req)
        return Rule._of(response.get_rule_res.rule) if response.get_rule_res.WhichOneof("res") == "rule" else None

    def get_rules(self):
        method = logic_proto.LogicManager.Req()
        method.get_rules_req.CopyFrom(logic_proto.LogicManager.GetRules.Req())
        return self._rule_stream(method, lambda res: res.get_rules_res.rules)

    def _execute(self, request: logic_proto.LogicManager.Req):
        req = transaction_proto.Transaction.Req()
        req.logic_manager_req.CopyFrom(request)
        return self._transaction._execute(req).logic_manager_res

    def _rule_stream(self, method: logic_proto.LogicManager.Req, rule_list_getter: Callable[[logic_proto.LogicManager.Res], List[logic_proto.Rule]]):
        request = transaction_proto.Transaction.Req()
        request.logic_manager_req.CopyFrom(method)
        return map(Rule._of, self._transaction._stream(request, lambda res: rule_list_getter(res.logic_manager_res)))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from datetime import datetime
from typing import List

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.common.exception import GraknClientException
from grakn.concept.type.value_type import ValueType


def iid(iid_: str):
    return bytes.fromhex(iid_.lstrip("0x"))


def thing(thing_):
    proto_thing = concept_proto.Thing()
    proto_thing.iid = iid(thing_.get_iid())
    proto_thing.encoding = thing_encoding(thing_)
    return proto_thing


def type_(_type):
    proto_type = concept_proto.Type()
    proto_type.label = _type.get_label()
    proto_type.encoding = type_encoding(_type)

    if _type.is_role_type():
        proto_type.scope = _type.get_scope()

    return proto_type


def types(types_: List):
    return map(lambda _type: type_(_type), types_)


def boolean_attribute_value(value: bool):
    value_proto = concept_proto.Attribute.Value()
    value_proto.boolean = value
    return value_proto


def long_attribute_value(value: int):
    value_proto = concept_proto.Attribute.Value()
    value_proto.long = value
    return value_proto


def double_attribute_value(value: float):
    value_proto = concept_proto.Attribute.Value()
    value_proto.double = value
    return value_proto


def string_attribute_value(value: str):
    value_proto = concept_proto.Attribute.Value()
    value_proto.string = value
    return value_proto


def datetime_attribute_value(value: datetime):
    value_proto = concept_proto.Attribute.Value()
    value_proto.date_time = int((value - datetime(1970, 1, 1)).total_seconds() * 1000)
    return value_proto


def value_type(value_type_: ValueType):
    if value_type_ == ValueType.BOOLEAN:
        return concept_proto.AttributeType.ValueType.Value("BOOLEAN")
    elif value_type_ == ValueType.LONG:
        return concept_proto.AttributeType.ValueType.Value("LONG")
    elif value_type_ == ValueType.DOUBLE:
        return concept_proto.AttributeType.ValueType.Value("DOUBLE")
    elif value_type_ == ValueType.STRING:
        return concept_proto.AttributeType.ValueType.Value("STRING")
    elif value_type_ == ValueType.DATETIME:
        return concept_proto.AttributeType.ValueType.Value("DATETIME")
    elif value_type_ == ValueType.OBJECT:
        return concept_proto.AttributeType.ValueType.Value("OBJECT")
    else:
        raise GraknClientException("Unrecognised value type: " + str(value_type_))


def thing_encoding(thing_):
    if thing_.is_entity():
        return concept_proto.Thing.Encoding.Value("ENTITY")
    elif thing_.is_relation():
        return concept_proto.Thing.Encoding.Value("RELATION")
    elif thing_.is_attribute():
        return concept_proto.Thing.Encoding.Value("ATTRIBUTE")
    else:
        raise GraknClientException("Unrecognised thing encoding: " + str(thing_))


def type_encoding(_type):
    if _type.is_entity_type():
        return concept_proto.Type.Encoding.Value("ENTITY_TYPE")
    elif _type.is_relation_type():
        return concept_proto.Type.Encoding.Value("RELATION_TYPE")
    elif _type.is_attribute_type():
        return concept_proto.Type.Encoding.Value("ATTRIBUTE_TYPE")
    elif _type.is_role_type():
        return concept_proto.Type.Encoding.Value("ROLE_TYPE")
    elif _type.is_thing_type():
        return concept_proto.Type.Encoding.Value("THING_TYPE")
    else:
        raise GraknClientException("Unrecognised type encoding: " + str(_type))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import abstractmethod
from typing import Dict, List

import grakn_protocol.protobuf.cluster.database_pb2 as database_proto

from grakn.common.exception import GraknClientException
from grakn.rpc.cluster.database import _DatabaseClusterRPC
from grakn.rpc.cluster.server_address import ServerAddress
from grakn.rpc.database import DatabaseCluster
from grakn.rpc.database_manager import DatabaseManager, _DatabaseManagerRPC


class DatabaseManagerCluster(DatabaseManager):

    @abstractmethod
    def get(self, name: str) -> DatabaseCluster:
        pass

    @abstractmethod
    def all(self) -> List[DatabaseCluster]:
        pass


class _DatabaseManagerClusterRPC(DatabaseManagerCluster):

    def __init__(self, client, database_managers: Dict[ServerAddress, "_DatabaseManagerRPC"]):
        self._client = client
        self._database_managers = database_managers

    def contains(self, name: str) -> bool:
        errors = []
        for database_manager in self._database_managers.values():
            try:
                return database_manager.contains(name)
            except GraknClientException as e:
                errors.append(e)
        raise GraknClientException("Attempted connecting to all cluster members, but the following errors occurred: " + str([str(e) for e in errors]))

    def create(self, name: str) -> None:
        for database_manager in self._database_managers.values():
            if not database_manager.contains(name):
                database_manager.create(name)

    def get(self, name: str) -> DatabaseCluster:
        errors = []
        for address in self._database_managers:
            try:
                database_get_req = database_proto.Database.Get.Req()
                database_get_req.name = name
                res = self._client.grakn_cluster_grpc_stub(address).database_get(database_get_req)
                return _DatabaseClusterRPC.of(res.database, self)
            except GraknClientException as e:
                errors.append(e)
        raise GraknClientException("Attempted connecting to all cluster members, but the following errors occurred: " + str([str(e) for e in errors]))

    def all(self) -> List[DatabaseCluster]:
        errors = []
        for address in self._database_managers:
            try:
                res = self._client.grakn_cluster_grpc_stub(address).database_all(database_proto.Database.All.Req())
                return [_DatabaseClusterRPC.of(db, self) for db in res.databases]
            except GraknClientException as e:
                errors.append(e)
        raise GraknClientException("Attempted connecting to all cluster members, but the following errors occurred: " + str([str(e) for e in errors]))

    def database_managers(self) -> Dict[ServerAddress, _DatabaseManagerRPC]:
        return self._database_managers



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Dict, Optional, Set

import grakn_protocol.protobuf.cluster.database_pb2 as database_proto

from grakn.rpc.cluster.server_address import ServerAddress
from grakn.rpc.database import DatabaseCluster, _DatabaseRPC


class _DatabaseClusterRPC(DatabaseCluster):

    def __init__(self, database_manager_cluster, database: str):
        self._databases: Dict[ServerAddress, _DatabaseRPC] = {}
        for address in database_manager_cluster.database_managers():
            database_manager = database_manager_cluster.database_managers()[address]
            self._databases[address] = _DatabaseRPC(database_manager, name=database)
        self._name = database
        self._database_manager_cluster = database_manager_cluster
        self._replicas: Set["_DatabaseClusterRPC.Replica"] = set()

    @staticmethod
    def of(proto_db: database_proto.Database, database_manager_cluster) -> "_DatabaseClusterRPC":
        assert proto_db.replicas
        database: str = proto_db.name
        database_cluster_rpc = _DatabaseClusterRPC(database_manager_cluster, database)
        for proto_replica in proto_db.replicas:
            database_cluster_rpc.replicas().add(_DatabaseClusterRPC.Replica.of(proto_replica, database_cluster_rpc))
        print("Discovered database cluster: " + str(database_cluster_rpc))
        return database_cluster_rpc

    def primary_replica(self) -> Optional["_DatabaseClusterRPC.Replica"]:
        primaries = [replica for replica in self._replicas if replica.is_primary()]
        return max(primaries, key=lambda r: r.term) if primaries else None

    def preferred_secondary_replica(self) -> "_DatabaseClusterRPC.Replica":
        return next(iter([replica for replica in self._replicas if replica.is_preferred_secondary()]), next(iter(self._replicas)))

    def name(self) -> str:
        return self._name

    def delete(self) -> None:
        for address in self._databases:
            if self._database_manager_cluster.database_managers()[address].contains(self._name):
                self._databases[address].delete()

    def replicas(self):
        return self._replicas

    def __str__(self):
        return self._name

    class Replica:

        def __init__(self, database: "_DatabaseClusterRPC", address: ServerAddress, term: int, is_primary: bool, is_preferred_secondary: bool):
            self._database = database
            self._replica_id = _DatabaseClusterRPC.Replica.Id(address, database.name())
            self._term = term
            self._is_primary = is_primary
            self._is_preferred_secondary = is_preferred_secondary

        @staticmethod
        def of(proto_replica: database_proto.Database.Replica, database: "_DatabaseClusterRPC") -> "_DatabaseClusterRPC.Replica":
            return _DatabaseClusterRPC.Replica(database, ServerAddress.parse(proto_replica.address), proto_replica.term,
                                               proto_replica.primary, proto_replica.preferred_secondary)

        def replica_id(self) -> "_DatabaseClusterRPC.Replica.Id":
            return self._replica_id

        def term(self) -> int:
            return self._term

        def is_primary(self) -> bool:
            return self._is_primary

        def is_preferred_secondary(self) -> bool:
            return self._is_preferred_secondary

        def address(self) -> ServerAddress:
            return self._replica_id.address()

        def __eq__(self, other):
            if self is other:
                return True
            if not other or type(self) != type(other):
                return False
            return self._replica_id == other.replica_id() and self._term == other.term() and self._is_primary == other.is_primary() and self._is_preferred_secondary == other.is_preferred_secondary()

        def __hash__(self):
            return hash((self._replica_id, self._is_primary, self._is_preferred_secondary, self._term))

        def __str__(self):
            return "%s:%s:%d" % (str(self._replica_id), "P" if self._is_primary else "S", self._term)

        class Id:

            def __init__(self, address: ServerAddress, database: str):
                self._address = address
                self._database = database

            def address(self) -> ServerAddress:
                return self._address

            def database(self) -> str:
                return self._database

            def __eq__(self, other):
                if self is other:
                    return True
                if not other or type(self) != type(other):
                    return False
                return self._address == other.address() and self._database == other.database()

            def __hash__(self):
                return hash((self._address, self._database))

            def __str__(self):
                return "%s/%s" % (str(self._address), self._database)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import ABC, abstractmethod
from typing import Optional


class GraknOptions(ABC):

    def __init__(self):
        self.infer: Optional[bool] = None
        self.trace_inference: Optional[bool] = None
        self.explain: Optional[bool] = None
        self.parallel: Optional[bool] = None
        self.batch_size: Optional[int] = None

    @staticmethod
    def core() -> "GraknOptions":
        return _GraknOptions()

    @staticmethod
    def cluster() -> "GraknClusterOptions":
        return _GraknClusterOptions()

    @abstractmethod
    def is_cluster(self) -> bool:
        pass


class GraknClusterOptions(GraknOptions, ABC):

    def __init__(self):
        super().__init__()
        self.read_any_replica: Optional[bool] = None


class _GraknOptions(GraknOptions):

    def is_cluster(self) -> bool:
        return False


class _GraknClusterOptions(GraknClusterOptions):

    def is_cluster(self) -> bool:
        return True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.answer_pb2 as answer_proto

from grakn.common.exception import GraknClientException


class Numeric:
    def __init__(self, int_value, float_value):
        self._int_value = int_value
        self._float_value = float_value

    def is_int(self):
        return self._int_value is not None

    def is_float(self):
        return self._float_value is not None

    def is_nan(self):
        return not self.is_int() and not self.is_float()

    def as_float(self):
        if (self.is_float()):
            return self._float_value
        else:
            raise GraknClientException("Illegal casting operation to 'float'.")

    def as_int(self):
        if (self.is_int()):
            return self._int_value
        else:
            raise GraknClientException("Illegal casting operation to 'int'.")


def _of(numeric_proto: answer_proto.Numeric):
    numeric_case = numeric_proto.WhichOneof("value")
    if numeric_case == "long_value":
        return Numeric(numeric_proto.long_value, None)
    elif numeric_case == "double_value":
        return Numeric(None, numeric_proto.double_value)
    elif numeric_case == "nan":
        return Numeric(None, None)
    else:
        raise GraknClientException("The answer type '" + numeric_case + "' was not recognised.")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
from abc import ABC, abstractmethod

import grakn_protocol.protobuf.cluster.database_pb2 as database_proto
from grpc import RpcError, StatusCode

from grakn.common.exception import GraknClientException
from grakn.rpc.cluster.database import _DatabaseClusterRPC


class _FailsafeTask(ABC):

    PRIMARY_REPLICA_TASK_MAX_RETRIES = 10
    FETCH_REPLICAS_MAX_RETRIES = 10
    WAIT_FOR_PRIMARY_REPLICA_SELECTION_SECONDS: float = 2

    def __init__(self, client, database: str):
        self.client = client
        self.database = database

    @abstractmethod
    def run(self, replica: _DatabaseClusterRPC.Replica):
        pass

    def rerun(self, replica: _DatabaseClusterRPC.Replica):
        return self.run(replica)

    def run_primary_replica(self):
        if self.database not in self.client.cluster_databases() or not self.client.cluster_databases()[self.database].primary_replica():
            self._seek_primary_replica()
        replica = self.client.cluster_databases()[self.database].primary_replica()
        retries = 0
        while True:
            try:
                return self.run(replica) if retries == 0 else self.rerun(replica)
            except GraknClientException as e:
                # TODO: propagate exception from the server in a less brittle way
                if "[RPL01]" in str(e):  # The contacted replica reported that it was not the primary replica
                    print("Unable to open a session or transaction, retrying in 2s... %s" % str(e))
                    time.sleep(self.WAIT_FOR_PRIMARY_REPLICA_SELECTION_SECONDS)
                    replica = self._seek_primary_replica()
                else:
                    raise e
                # TODO: introduce a special type that extends RpcError and Call
            except RpcError as e:
                # TODO: this logic should be extracted into GraknClientException
                # TODO: error message should be checked in a less brittle way
                if e.code() == StatusCode.UNAVAILABLE or "[INT07]" in str(e) or "Received RST_STREAM" in str(e):
                    print("Unable to open a session or transaction, retrying in 2s... %s" % str(e))
                    time.sleep(self.WAIT_FOR_PRIMARY_REPLICA_SELECTION_SECONDS)
                    replica = self._seek_primary_replica()
                else:
                    raise e
            retries += 1
            if retries > self.PRIMARY_REPLICA_TASK_MAX_RETRIES:
                raise self._cluster_not_available_exception()

    def run_any_replica(self):
        if self.database in self.client.cluster_databases():
            replica_info = self.client.cluster_databases()[self.database]
        else:
            replica_info = self._fetch_database_replicas()

        replicas = [replica_info.preferred_secondary_replica()] + [replica for replica in replica_info.replicas() if not replica.is_preferred_secondary()]
        retries = 0
        for replica in replicas:
            try:
                return self.run(replica) if retries == 0 else self.rerun(replica)
            except RpcError as e:
                if e.code() == StatusCode.UNAVAILABLE or "[INT07]" in str(e) or "Received RST_STREAM" in str(e):
                    print("Unable to open a session or transaction to %s. Attempting next replica. %s" % (str(replica.replica_id()), str(e)))
                else:
                    raise e
            retries += 1
        raise self._cluster_not_available_exception()

    def _seek_primary_replica(self) -> _DatabaseClusterRPC.Replica:
        retries = 0
        while retries < self.FETCH_REPLICAS_MAX_RETRIES:
            replica_info = self._fetch_database_replicas()
            if replica_info.primary_replica():
                return replica_info.primary_replica()
            else:
                time.sleep(self.WAIT_FOR_PRIMARY_REPLICA_SELECTION_SECONDS)
                retries += 1
        raise self._cluster_not_available_exception()

    def _fetch_database_replicas(self) -> _DatabaseClusterRPC:
        for server_address in self.client.cluster_members():
            try:
                print("Fetching replica info from %s" % server_address)
                db_get_req = database_proto.Database.Get.Req()
                db_get_req.name = self.database
                res = self.client.grakn_cluster_grpc_stub(server_address).database_get(db_get_req)
                replica_info = _DatabaseClusterRPC.of(res.database, self.client.databases())
                print("Requested database discovery from peer %s, and got response: %s" % (str(server_address), str([str(replica) for replica in replica_info.replicas()])))
                self.client.cluster_databases()[self.database] = replica_info
                return replica_info
            except RpcError as e:
                print("Unable to perform database discovery to %s. Attempting next address. %s" % (str(server_address), str(e)))
        raise self._cluster_not_available_exception()

    def _cluster_not_available_exception(self) -> GraknClientException:
        addresses = str([str(addr) for addr in self.client.cluster_members()])
        return GraknClientException("Unable to connect to Grakn Cluster. Attempted connecting to the cluster members, but none are available: '%s'" % addresses)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from datetime import datetime
from typing import Optional

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.common.exception import GraknClientException
from grakn.concept.proto import concept_proto_builder, concept_proto_reader
from grakn.concept.type.thing_type import ThingType, RemoteThingType
from grakn.concept.type.value_type import ValueType


def is_keyable(value_type: ValueType):
    return value_type in [ValueType.LONG, ValueType.STRING, ValueType.DATETIME]


class AttributeType(ThingType):

    ROOT_LABEL = "attribute"

    def as_remote(self, transaction):
        return RemoteAttributeType(transaction, self.get_label(), self.is_root())

    def get_value_type(self):
        return ValueType.OBJECT

    def is_keyable(self):
        return is_keyable(self.get_value_type())

    def is_attribute_type(self):
        return True

    def is_boolean(self):
        return False

    def is_long(self):
        return False

    def is_double(self):
        return False

    def is_string(self):
        return False

    def is_datetime(self):
        return False

    def as_boolean(self):
        if self.is_root():
            return BooleanAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + BooleanAttributeType.__name__)

    def as_long(self):
        if self.is_root():
            return LongAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + LongAttributeType.__name__)

    def as_double(self):
        if self.is_root():
            return DoubleAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + DoubleAttributeType.__name__)

    def as_string(self):
        if self.is_root():
            return StringAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + StringAttributeType.__name__)

    def as_datetime(self):
        if self.is_root():
            return DateTimeAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + DateTimeAttributeType.__name__)

    def __eq__(self, other):
        if other is self:
            return True
        # root "attribute" should always be equal to itself regardless of which value class it holds
        if not other or not isinstance(other, AttributeType):
            return False
        return self.get_label() == other.get_label()

    def __hash__(self):
        return super(AttributeType, self).__hash__()


class RemoteAttributeType(RemoteThingType):

    ROOT_LABEL = "attribute"

    def get_value_type(self):
        return ValueType.OBJECT

    def is_keyable(self):
        return is_keyable(self.get_value_type())

    def as_remote(self, transaction):
        return RemoteAttributeType(transaction, self.get_label(), self.is_root())

    def get_subtypes(self):
        stream = super(RemoteAttributeType, self).get_subtypes()
        if self.is_root() and self.get_value_type() is not ValueType.OBJECT:
            return [subtype for subtype in stream if subtype.get_value_type() is self.get_value_type() or subtype.get_label() == self.get_label()]
        else:
            return stream

    def get_owners(self, only_key=False):
        method = concept_proto.Type.Req()
        get_owners_req = concept_proto.AttributeType.GetOwners.Req()
        get_owners_req.only_key = only_key
        method.attribute_type_get_owners_req.CopyFrom(get_owners_req)
        return self._type_stream(method, lambda res: res.attribute_type_get_owners_res.owners)

    def _put_internal(self, value_proto):
        method = concept_proto.Type.Req()
        put_req = concept_proto.AttributeType.Put.Req()
        put_req.value.CopyFrom(value_proto)
        method.attribute_type_put_req.CopyFrom(put_req)
        return concept_proto_reader.attribute(self._execute(method).attribute_type_put_res.attribute)

    def _get_internal(self, value_proto):
        method = concept_proto.Type.Req()
        get_req = concept_proto.AttributeType.Get.Req()
        get_req.value.CopyFrom(value_proto)
        method.attribute_type_get_req.CopyFrom(get_req)
        response = self._execute(method).attribute_type_get_res
        return concept_proto_reader.attribute(response.attribute) if response.WhichOneof("res") == "attribute" else None

    def is_attribute_type(self):
        return True

    def is_boolean(self):
        return False

    def is_long(self):
        return False

    def is_double(self):
        return False

    def is_string(self):
        return False

    def is_datetime(self):
        return False

    def as_boolean(self):
        if self.is_root():
            return BooleanAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + BooleanAttributeType.__name__)

    def as_long(self):
        if self.is_root():
            return LongAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + LongAttributeType.__name__)

    def as_double(self):
        if self.is_root():
            return DoubleAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + DoubleAttributeType.__name__)

    def as_string(self):
        if self.is_root():
            return StringAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + StringAttributeType.__name__)

    def as_datetime(self):
        if self.is_root():
            return DateTimeAttributeType(self.ROOT_LABEL, is_root=True)
        raise GraknClientException("Invalid conversion to " + DateTimeAttributeType.__name__)

    def __eq__(self, other):
        if other is self:
            return True
        if not other or not isinstance(RemoteAttributeType, other):
            return False
        return self.get_label() == other.get_label()

    def __hash__(self):
        return super(RemoteAttributeType, self).__hash__()


class BooleanAttributeType(AttributeType):

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return BooleanAttributeType(type_proto.label, type_proto.root)

    def get_value_type(self):
        return ValueType.BOOLEAN

    def as_remote(self, transaction):
        return RemoteBooleanAttributeType(transaction, self.get_label(), self.is_root())

    def is_boolean(self):
        return True

    def as_boolean(self):
        return self


class RemoteBooleanAttributeType(RemoteAttributeType):

    def get_value_type(self):
        return ValueType.BOOLEAN

    def as_remote(self, transaction):
        return RemoteBooleanAttributeType(transaction, self.get_label(), self.is_root())

    def put(self, value: bool):
        return self._put_internal(concept_proto_builder.boolean_attribute_value(value))

    def get(self, value: bool):
        return self._get_internal(concept_proto_builder.boolean_attribute_value(value))

    def is_boolean(self):
        return True

    def as_boolean(self):
        return self


class LongAttributeType(AttributeType):

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return LongAttributeType(type_proto.label, type_proto.root)

    def get_value_type(self):
        return ValueType.LONG

    def as_remote(self, transaction):
        return RemoteLongAttributeType(transaction, self.get_label(), self.is_root())

    def is_long(self):
        return True

    def as_long(self):
        return self


class RemoteLongAttributeType(RemoteAttributeType):

    def get_value_type(self):
        return ValueType.LONG

    def as_remote(self, transaction):
        return RemoteLongAttributeType(transaction, self.get_label(), self.is_root())

    def put(self, value: int):
        return self._put_internal(concept_proto_builder.long_attribute_value(value))

    def get(self, value: int):
        return self._get_internal(concept_proto_builder.long_attribute_value(value))

    def is_long(self):
        return True

    def as_long(self):
        return self


class DoubleAttributeType(AttributeType):

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return DoubleAttributeType(type_proto.label, type_proto.root)

    def get_value_type(self):
        return ValueType.DOUBLE

    def as_remote(self, transaction):
        return RemoteDoubleAttributeType(transaction, self.get_label(), self.is_root())

    def is_double(self):
        return True

    def as_double(self):
        return self


class RemoteDoubleAttributeType(RemoteAttributeType):

    def get_value_type(self):
        return ValueType.DOUBLE

    def as_remote(self, transaction):
        return RemoteDoubleAttributeType(transaction, self.get_label(), self.is_root())

    def put(self, value: float):
        return self._put_internal(concept_proto_builder.double_attribute_value(value))

    def get(self, value: float):
        return self._get_internal(concept_proto_builder.double_attribute_value(value))

    def is_double(self):
        return True

    def as_double(self):
        return self


class StringAttributeType(AttributeType):

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return StringAttributeType(type_proto.label, type_proto.root)

    def get_value_type(self):
        return ValueType.STRING

    def as_remote(self, transaction):
        return RemoteStringAttributeType(transaction, self.get_label(), self.is_root())

    def is_string(self):
        return True

    def as_string(self):
        return self


class RemoteStringAttributeType(RemoteAttributeType):

    def get_value_type(self):
        return ValueType.STRING

    def as_remote(self, transaction):
        return RemoteStringAttributeType(transaction, self.get_label(), self.is_root())

    def put(self, value: str):
        return self._put_internal(concept_proto_builder.string_attribute_value(value))

    def get(self, value: str):
        return self._get_internal(concept_proto_builder.string_attribute_value(value))

    def get_regex(self):
        method = concept_proto.Type.Req()
        get_regex_req = concept_proto.AttributeType.GetRegex.Req()
        method.attribute_type_get_regex_req.CopyFrom(get_regex_req)
        regex = self._execute(method).attribute_type_get_regex_res.regex
        return None if len(regex) == 0 else regex

    def set_regex(self, regex: Optional[str]):
        if regex is None:
            regex = ""
        method = concept_proto.Type.Req()
        set_regex_req = concept_proto.AttributeType.SetRegex.Req()
        set_regex_req.regex = regex
        method.attribute_type_set_regex_req.CopyFrom(set_regex_req)
        self._execute(method)

    def is_string(self):
        return True

    def as_string(self):
        return self


class DateTimeAttributeType(AttributeType):

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return DateTimeAttributeType(type_proto.label, type_proto.root)

    def get_value_type(self):
        return ValueType.DATETIME

    def as_remote(self, transaction):
        return RemoteDateTimeAttributeType(transaction, self.get_label(), self.is_root())

    def is_datetime(self):
        return True

    def as_datetime(self):
        return self


class RemoteDateTimeAttributeType(RemoteAttributeType):

    def get_value_type(self):
        return ValueType.DATETIME

    def as_remote(self, transaction):
        return RemoteDateTimeAttributeType(transaction, self.get_label(), self.is_root())

    def put(self, value: datetime):
        return self._put_internal(concept_proto_builder.datetime_attribute_value(value))

    def get(self, value: datetime):
        return self._get_internal(concept_proto_builder.datetime_attribute_value(value))

    def is_datetime(self):
        return True

    def as_datetime(self):
        return self



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.answer_pb2 as answer_proto
from grakn.concept.answer import concept_map
from grakn.concept.proto.concept_proto_reader import concept

class ConceptMapGroup:
    def __init__(self, owner, concept_maps):
        self._owner = owner
        self._concept_maps = concept_maps

    def owner(self):
        return self._owner

    def concept_maps(self):
        return self._concept_maps


def _of(concept_map_group_proto: answer_proto.ConceptMapGroup):
    return ConceptMapGroup(concept(concept_map_group_proto.owner), map(lambda cm: concept_map._of(cm), concept_map_group_proto.concept_maps))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import enum
import sched
import time
from abc import ABC, abstractmethod
from threading import Thread, Lock

import grakn_protocol.protobuf.session_pb2 as session_proto
import grpc
from grakn_protocol.protobuf.grakn_pb2_grpc import GraknStub

from grakn import grakn_proto_builder
from grakn.options import GraknOptions
from grakn.rpc.database import Database, _DatabaseRPC
from grakn.rpc.transaction import Transaction, TransactionType


class SessionType(enum.Enum):
    DATA = 0
    SCHEMA = 1


def _session_type_proto(session_type: SessionType):
    if session_type == SessionType.DATA:
        return session_proto.Session.Type.Value("DATA")
    if session_type == SessionType.SCHEMA:
        return session_proto.Session.Type.Value("SCHEMA")


class Session(ABC):

    @abstractmethod
    def transaction(self, transaction_type: TransactionType, options=None) -> Transaction:
        pass

    @abstractmethod
    def session_type(self) -> SessionType:
        pass

    @abstractmethod
    def is_open(self) -> bool:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def database(self) -> Database:
        pass

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class _SessionRPC(Session):
    _PULSE_FREQUENCY_SECONDS = 5

    def __init__(self, client, database: str, session_type: SessionType, options: GraknOptions = None):
        if not options:
            options = GraknOptions.core()
        self._client = client
        self._address = client._address
        self._channel = grpc.insecure_channel(client._address)
        self._scheduler = sched.scheduler(time.time, time.sleep)
        self._database = _DatabaseRPC(database_manager=client.databases(), name=database)
        self._session_type = session_type
        self._grpc_stub = GraknStub(self._channel)
        self._lock = Lock()

        open_req = session_proto.Session.Open.Req()
        open_req.database = database
        open_req.type = _session_type_proto(session_type)
        open_req.options.CopyFrom(grakn_proto_builder.options(options))

        self._session_id: bytes = self._grpc_stub.session_open(open_req).session_id
        self._is_open = True
        self._pulse = self._scheduler.enter(delay=self._PULSE_FREQUENCY_SECONDS, priority=1, action=self._transmit_pulse, argument=())
        Thread(target=self._scheduler.run, name="session_pulse_{}".format(self._session_id.hex()), daemon=True).start()

    def transaction(self, transaction_type: TransactionType, options=None) -> Transaction:
        if not options:
            options = GraknOptions.core()
        return Transaction(self._address, self._session_id, transaction_type, options)

    def session_type(self) -> SessionType:
        return self._session_type

    def is_open(self) -> bool:
        return self._is_open

    def close(self) -> None:
        self._lock.acquire(blocking=True)
        if self._is_open:
            self._is_open = False
            self._client.remove_session(self)
            self._lock.release()
            try:
                self._scheduler.cancel(self._pulse)
            except ValueError:  # This may occur if a pulse is in transit but not responded to yet.
                pass
            req = session_proto.Session.Close.Req()
            req.session_id = self._session_id
            try:
                self._grpc_stub.session_close(req)
            finally:
                self._channel.close()
        else:
            self._lock.release()

    def database(self) -> Database:
        return self._database

    def session_id(self) -> bytes:
        return self._session_id

    def _transmit_pulse(self) -> None:
        if not self._is_open:
            return
        pulse_req = session_proto.Session.Pulse.Req()
        pulse_req.session_id = self._session_id
        res = self._grpc_stub.session_pulse(pulse_req)
        if res.alive:
            self._pulse = self._scheduler.enter(delay=self._PULSE_FREQUENCY_SECONDS, priority=1, action=self._transmit_pulse, argument=())
            Thread(target=self._scheduler.run, name="session_pulse_{}".format(self._session_id.hex()), daemon=True).start()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_tb is None:
            pass
        else:
            return False



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from setuptools import setup
from setuptools import find_packages

packages = find_packages()

setup(
    name = "grakn-client",
    version = "2.0.0-alpha-9",
    description = "Grakn Client for Python",
    long_description = open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers = ["Programming Language :: Python", "Programming Language :: Python :: 3", "Programming Language :: Python :: 3.5", "Programming Language :: Python :: 3.6", "Programming Language :: Python :: 3.7", "Programming Language :: Python :: 3.8", "License :: OSI Approved :: Apache Software License", "Operating System :: OS Independent", "Intended Audience :: Developers", "Intended Audience :: Science/Research", "Environment :: Console", "Topic :: Database :: Front-Ends"],
    keywords = "grakn database graph knowledgebase knowledge-engineering",
    url = "https://github.com/graknlabs/client-python/",
    author = "Grakn Labs",
    author_email = "community@grakn.ai",
    license = "Apache-2.0",
    packages=packages,
    install_requires=['grakn-protocol==2.0.0a11', 'grpcio==1.35.0', 'protobuf==3.14.0'],
    zip_safe=False,
)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.concept_pb2 as concept_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.concept.proto import concept_proto_reader, concept_proto_builder
from grakn.concept.type.entity_type import EntityType
from grakn.concept.type.relation_type import RelationType


class ConceptManager:

    def __init__(self, transaction):
        self._transaction = transaction

    def get_root_thing_type(self):
        return self.get_thing_type("thing")

    def get_root_entity_type(self):
        return self.get_entity_type("entity")

    def get_root_relation_type(self):
        return self.get_relation_type("relation")

    def get_root_attribute_type(self):
        return self.get_attribute_type("attribute")

    def put_entity_type(self, label: str):
        req = concept_proto.ConceptManager.Req()
        put_entity_type_req = concept_proto.ConceptManager.PutEntityType.Req()
        put_entity_type_req.label = label
        req.put_entity_type_req.CopyFrom(put_entity_type_req)
        res = self._execute(req)
        return EntityType._of(res.put_entity_type_res.entity_type)

    def get_entity_type(self, label: str):
        _type = self.get_thing_type(label)
        return _type if _type and _type.is_entity_type() else None

    def put_relation_type(self, label: str):
        req = concept_proto.ConceptManager.Req()
        put_relation_type_req = concept_proto.ConceptManager.PutRelationType.Req()
        put_relation_type_req.label = label
        req.put_relation_type_req.CopyFrom(put_relation_type_req)
        res = self._execute(req)
        return RelationType._of(res.put_relation_type_res.relation_type)

    def get_relation_type(self, label: str):
        _type = self.get_thing_type(label)
        return _type if _type and _type.is_relation_type() else None

    def put_attribute_type(self, label: str, value_type):
        req = concept_proto.ConceptManager.Req()
        put_attribute_type_req = concept_proto.ConceptManager.PutAttributeType.Req()
        put_attribute_type_req.label = label
        put_attribute_type_req.value_type = concept_proto_builder.value_type(value_type)
        req.put_attribute_type_req.CopyFrom(put_attribute_type_req)
        res = self._execute(req)
        return concept_proto_reader.attribute_type(res.put_attribute_type_res.attribute_type)

    def get_attribute_type(self, label: str):
        _type = self.get_thing_type(label)
        return _type if _type and _type.is_attribute_type() else None

    def get_thing(self, iid: str):
        req = concept_proto.ConceptManager.Req()
        get_thing_req = concept_proto.ConceptManager.GetThing.Req()
        get_thing_req.iid = concept_proto_builder.iid(iid)
        req.get_thing_req.CopyFrom(get_thing_req)

        response = self._execute(req)
        return concept_proto_reader.thing(response.get_thing_res.thing) if response.get_thing_res.WhichOneof("res") == "thing" else None

    def get_thing_type(self, label: str):
        req = concept_proto.ConceptManager.Req()
        get_thing_type_req = concept_proto.ConceptManager.GetThingType.Req()
        get_thing_type_req.label = label
        req.get_thing_type_req.CopyFrom(get_thing_type_req)

        response = self._execute(req)
        return concept_proto_reader.thing_type(response.get_thing_type_res.thing_type) if response.get_thing_type_res.WhichOneof("res") == "thing_type" else None

    def _execute(self, request: concept_proto.ConceptManager.Req):
        req = transaction_proto.Transaction.Req()
        req.concept_manager_req.CopyFrom(request)
        return self._transaction._execute(req).concept_manager_res



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.logic_pb2 as logic_proto
import grakn_protocol.protobuf.transaction_pb2 as transaction_proto

from grakn.common.exception import GraknClientException


class Rule:

    def __init__(self, label: str, when: str, then: str):
        if not label:
            raise GraknClientException("Label must be a non-empty string.")
        self._label = label
        self._when = when
        self._then = then
        self._hash = hash(label)

    @staticmethod
    def _of(rule_proto: logic_proto.Rule):
        return Rule(rule_proto.label, rule_proto.when, rule_proto.then)

    def get_label(self):
        return self._label

    def get_when(self):
        return self._when

    def get_then(self):
        return self._then

    def as_remote(self, transaction):
        return RemoteRule(transaction, self.get_label(), self.get_when(), self.get_then())

    def is_remote(self):
        return False

    def __str__(self):
        return type(self).__name__ + "[label:" + self.get_label() + "]"

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self.get_label() == other.get_label()

    def __hash__(self):
        return self._hash


class RemoteRule(Rule):

    def __init__(self, transaction, label: str, when: str, then: str):
        super(RemoteRule, self).__init__(label, when, then)
        if not transaction:
            raise GraknClientException("Transaction must be set.")
        self._transaction = transaction
        self._hash = hash((transaction, label))

    def set_label(self, label: str):
        req = logic_proto.Rule.Req()
        set_label_req = logic_proto.Rule.SetLabel.Req()
        set_label_req.label = label
        req.rule_set_label_req.CopyFrom(set_label_req)
        self._execute(req)
        self._label = label

    def delete(self):
        method = logic_proto.Rule.Req()
        method.rule_delete_req.CopyFrom(logic_proto.Rule.Delete.Req())
        self._execute(method)

    def is_deleted(self):
        return not self._transaction.logic().get_rule(self.get_label())

    def is_remote(self):
        return True

    def __eq__(self, other):
        if other is self:
            return True
        if not other or type(self) != type(other):
            return False
        return self._transaction is other._transaction and self.get_label() == other.get_label()

    def __hash__(self):
        return super(RemoteRule, self).__hash__()

    def _execute(self, method: logic_proto.Rule.Req):
        method.label = self.get_label()
        request = transaction_proto.Transaction.Req()
        request.rule_req.CopyFrom(method)
        return self._transaction._execute(request).rule_res



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Union

import grakn_protocol.protobuf.options_pb2 as options_proto

from grakn.options import GraknOptions, GraknClusterOptions


def options(opts: Union[GraknOptions, GraknClusterOptions]):
    proto_options = options_proto.Options()
    if opts.infer is not None:
        proto_options.infer = opts.infer
    if opts.trace_inference is not None:
        proto_options.trace_inference = opts.trace_inference
    if opts.explain is not None:
        proto_options.explain = opts.explain
    if opts.parallel is not None:
        proto_options.parallel = opts.parallel
    if opts.batch_size is not None:
        proto_options.batch_size = opts.batch_size
    if opts.is_cluster() and opts.read_any_replica is not None:
        proto_options.read_any_replica = opts.read_any_replica
    return proto_options



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.concept.proto import concept_proto_reader
from grakn.concept.thing.relation import Relation
from grakn.concept.type.thing_type import ThingType, RemoteThingType


class RelationType(ThingType):

    @staticmethod
    def _of(type_proto: concept_proto.Type):
        return RelationType(type_proto.label, type_proto.root)

    def as_remote(self, transaction):
        return RemoteRelationType(transaction, self.get_label(), self.is_root())

    def is_relation_type(self):
        return True


class RemoteRelationType(RemoteThingType):

    def as_remote(self, transaction):
        return RemoteRelationType(transaction, self.get_label(), self.is_root())

    def create(self):
        method = concept_proto.Type.Req()
        create_req = concept_proto.RelationType.Create.Req()
        method.relation_type_create_req.CopyFrom(create_req)
        return Relation._of(self._execute(method).relation_type_create_res.relation)

    def get_relates(self, role_label: str = None):
        method = concept_proto.Type.Req()
        if role_label:
            get_relates_req = concept_proto.RelationType.GetRelatesForRoleLabel.Req()
            get_relates_req.label = role_label
            method.relation_type_get_relates_for_role_label_req.CopyFrom(get_relates_req)
            res = self._execute(method).relation_type_get_relates_for_role_label_res
            return concept_proto_reader.type_(res.role_type) if res.HasField("role_type") else None
        else:
            method.relation_type_get_relates_req.CopyFrom(concept_proto.RelationType.GetRelates.Req())
            return self._type_stream(method, lambda res: res.relation_type_get_relates_res.roles)

    def set_relates(self, role_label: str, overridden_label: str = None):
        method = concept_proto.Type.Req()
        set_relates_req = concept_proto.RelationType.SetRelates.Req()
        set_relates_req.label = role_label
        if overridden_label:
            set_relates_req.overridden_label = overridden_label
        method.relation_type_set_relates_req.CopyFrom(set_relates_req)
        self._execute(method)

    def unset_relates(self, role_label: str):
        method = concept_proto.Type.Req()
        unset_relates_req = concept_proto.RelationType.UnsetRelates.Req()
        unset_relates_req.label = role_label
        method.relation_type_unset_relates_req.CopyFrom(unset_relates_req)
        self._execute(method)

    def is_relation_type(self):
        return True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from grakn.options import GraknClusterOptions, GraknOptions
from grakn.rpc.cluster.failsafe_task import _FailsafeTask
from grakn.rpc.cluster.database import _DatabaseClusterRPC
from grakn.rpc.cluster.server_address import ServerAddress
from grakn.rpc.database import Database
from grakn.rpc.session import Session, SessionType
from grakn.rpc.transaction import TransactionType, Transaction


class SessionClusterRPC(Session):

    def __init__(self, cluster_client, server_address: ServerAddress, database: str, session_type: SessionType, options: GraknClusterOptions):
        self.cluster_client = cluster_client
        self.core_client = cluster_client.core_client(server_address)
        print("Opening a session to %s" % server_address)
        self.core_session = self.core_client.session(database, session_type, options)

    def transaction(self, transaction_type: TransactionType, options: GraknClusterOptions = None) -> Transaction:
        if not options:
            options = GraknOptions.cluster()
        return self._transaction_any_replica(transaction_type, options) if options.read_any_replica else self._transaction_primary_replica(transaction_type, options)

    def _transaction_primary_replica(self, transaction_type: TransactionType, options: GraknClusterOptions) -> Transaction:
        return TransactionFailsafeTask(self, transaction_type, options).run_primary_replica()

    def _transaction_any_replica(self, transaction_type: TransactionType, options: GraknClusterOptions) -> Transaction:
        return TransactionFailsafeTask(self, transaction_type, options).run_any_replica()

    def session_type(self) -> SessionType:
        return self.core_session.session_type()

    def is_open(self) -> bool:
        return self.core_session.is_open()

    def close(self) -> None:
        self.core_session.close()

    def database(self) -> Database:
        return self.core_session.database()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TransactionFailsafeTask(_FailsafeTask):

    def __init__(self, cluster_session: SessionClusterRPC, transaction_type: TransactionType, options: GraknClusterOptions):
        super().__init__(cluster_session.cluster_client, cluster_session.database().name())
        self.cluster_session = cluster_session
        self.transaction_type = transaction_type
        self.options = options

    def run(self, replica: _DatabaseClusterRPC.Replica):
        return self.cluster_session.core_session.transaction(self.transaction_type, self.options)

    def rerun(self, replica: _DatabaseClusterRPC.Replica):
        if self.cluster_session.core_session:
            self.cluster_session.core_session.close()
        self.cluster_session.core_client = self.cluster_session.cluster_client.core_client(replica.address())
        self.cluster_session.core_session = self.cluster_session.core_client.session(self.database, self.cluster_session.session_type(), self.options)
        return self.cluster_session.core_session.transaction(self.transaction_type, self.options)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from datetime import datetime

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.concept.proto import concept_proto_builder, concept_proto_reader
from grakn.concept.thing.thing import Thing, RemoteThing


class Attribute(Thing):

    def is_attribute(self):
        return True

    def is_boolean(self):
        return False

    def is_long(self):
        return False

    def is_double(self):
        return False

    def is_string(self):
        return False

    def is_datetime(self):
        return False


class RemoteAttribute(RemoteThing):

    def get_owners(self, owner_type=None):
        method = concept_proto.Thing.Req()
        get_owners_req = concept_proto.Attribute.GetOwners.Req()
        if owner_type:
            get_owners_req.thing_type = concept_proto_builder.type_(owner_type)
        method.attribute_get_owners_req.CopyFrom(get_owners_req)
        return self._thing_stream(method, lambda res: res.attribute_get_owners_res.things)

    def is_attribute(self):
        return True

    def is_boolean(self):
        return False

    def is_long(self):
        return False

    def is_double(self):
        return False

    def is_string(self):
        return False

    def is_datetime(self):
        return False


class BooleanAttribute(Attribute):

    def __init__(self, iid: str, value: bool):
        super(BooleanAttribute, self).__init__(iid)
        self._value = value

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return BooleanAttribute(concept_proto_reader.iid(thing_proto.iid), thing_proto.value.boolean)

    def get_value(self):
        return self._value

    def is_boolean(self):
        return True

    def as_remote(self, transaction):
        return RemoteBooleanAttribute(transaction, self.get_iid(), self.get_value())


class RemoteBooleanAttribute(RemoteAttribute):

    def __init__(self, transaction, iid: str, value: bool):
        super(RemoteBooleanAttribute, self).__init__(transaction, iid)
        self._value = value

    def get_value(self):
        return self._value

    def is_boolean(self):
        return True

    def as_remote(self, transaction):
        return RemoteBooleanAttribute(transaction, self.get_iid(), self.get_value())


class LongAttribute(Attribute):

    def __init__(self, iid: str, value: int):
        super(LongAttribute, self).__init__(iid)
        self._value = value

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return LongAttribute(concept_proto_reader.iid(thing_proto.iid), thing_proto.value.long)

    def get_value(self):
        return self._value

    def is_long(self):
        return True

    def as_remote(self, transaction):
        return RemoteLongAttribute(transaction, self.get_iid(), self.get_value())


class RemoteLongAttribute(RemoteAttribute):

    def __init__(self, transaction, iid: str, value: int):
        super(RemoteLongAttribute, self).__init__(transaction, iid)
        self._value = value

    def get_value(self):
        return self._value

    def is_long(self):
        return True

    def as_remote(self, transaction):
        return RemoteLongAttribute(transaction, self.get_iid(), self.get_value())


class DoubleAttribute(Attribute):

    def __init__(self, iid: str, value: float):
        super(DoubleAttribute, self).__init__(iid)
        self._value = value

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return DoubleAttribute(concept_proto_reader.iid(thing_proto.iid), thing_proto.value.double)

    def get_value(self):
        return self._value

    def is_double(self):
        return True

    def as_remote(self, transaction):
        return RemoteDoubleAttribute(transaction, self.get_iid(), self.get_value())


class RemoteDoubleAttribute(RemoteAttribute):

    def __init__(self, transaction, iid: str, value: float):
        super(RemoteDoubleAttribute, self).__init__(transaction, iid)
        self._value = value

    def get_value(self):
        return self._value

    def is_double(self):
        return True

    def as_remote(self, transaction):
        return RemoteDoubleAttribute(transaction, self.get_iid(), self.get_value())


class StringAttribute(Attribute):

    def __init__(self, iid: str, value: str):
        super(StringAttribute, self).__init__(iid)
        self._value = value

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return StringAttribute(concept_proto_reader.iid(thing_proto.iid), thing_proto.value.string)

    def get_value(self):
        return self._value

    def is_string(self):
        return True

    def as_remote(self, transaction):
        return RemoteStringAttribute(transaction, self.get_iid(), self.get_value())


class RemoteStringAttribute(RemoteAttribute):

    def __init__(self, transaction, iid: str, value: str):
        super(RemoteStringAttribute, self).__init__(transaction, iid)
        self._value = value

    def get_value(self):
        return self._value

    def is_string(self):
        return True

    def as_remote(self, transaction):
        return RemoteStringAttribute(transaction, self.get_iid(), self.get_value())


class DateTimeAttribute(Attribute):

    def __init__(self, iid: str, value: datetime):
        super(DateTimeAttribute, self).__init__(iid)
        self._value = value

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return DateTimeAttribute(concept_proto_reader.iid(thing_proto.iid), datetime.fromtimestamp(float(thing_proto.value.date_time) / 1000.0))

    def get_value(self):
        return self._value

    def is_datetime(self):
        return True

    def as_remote(self, transaction):
        return RemoteDateTimeAttribute(transaction, self.get_iid(), self.get_value())


class RemoteDateTimeAttribute(RemoteAttribute):

    def __init__(self, transaction, iid: str, value: datetime):
        super(RemoteDateTimeAttribute, self).__init__(transaction, iid)
        self._value = value

    def get_value(self):
        return self._value

    def is_datetime(self):
        return True

    def as_remote(self, transaction):
        return RemoteDateTimeAttribute(transaction, self.get_iid(), self.get_value())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import enum

from datetime import datetime
from typing import Union


class ValueType(enum.Enum):
    # This lives here to avoid circular imports.
    OBJECT = 0
    BOOLEAN = 1
    LONG = 2
    DOUBLE = 3
    STRING = 4
    DATETIME = 5


ValueClass = Union[bool, int, float, str, datetime]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import grakn_protocol.protobuf.concept_pb2 as concept_proto

from grakn.concept.proto import concept_proto_reader
from grakn.concept.thing.thing import Thing, RemoteThing


class Entity(Thing):

    @staticmethod
    def _of(thing_proto: concept_proto.Thing):
        return Entity(concept_proto_reader.iid(thing_proto.iid))

    def as_remote(self, transaction):
        return RemoteEntity(transaction, self._iid)

    def is_entity(self):
        return True


class RemoteEntity(RemoteThing):

    def as_remote(self, transaction):
        return RemoteEntity(transaction, self._iid)

    def is_entity(self):
        return True
